//-----------------------------------------------------------------------------
//
//  Title: TIA Audio Generator
//
//  Project - MMDC
//  Version - _VERSION
//  Author - Chris Brenner
//  Description - This code is translated more or less verbatim from my Verilog
//  code for my FPGA project. It provides core logic that can be used to achieve
//  cycle accurate emulation of the Atari 2600 TIA audio blocks.
//
//  The core logic is broken up into two functions. The update_aud_state()
//  function contains the logic for the clock divider and pulse generator. It is
//  used to update the state of the audio logic, and should be called once for
//  every audio clock cycle for each channel.
//
//  The create_sample() function implements the volume control, and generates
//  audio samples. It depends on the state of the pulse generator of both audio
//  channels, so it should be called after the update_aud_state() function calls.
//
//  The init_aud() function should be called once on startup in order to initialize
//  the audio logic and volume control LUT.
//
//  The accuracy of the emulation is dependent upon the accuracy of the data
//  contained in the AUDCx, AUDFx, and AUDVx registers. It is important that these
//  registers contain the most recent value written prior to the current clock
//  cycle and phase.
//
//  The TIA audio clock is a 31.4 KHz two phase clock that occurs twice every
//  horizontal scan line. The mapping of the cycles and phases are as follows.
//  Cycle 1 - Phase 1: color clock 8
//  Cycle 1 - Phase 2: color clock 36
//  Cycle 2 - Phase 1: color clock 80
//  Cycle 2 - Phase 2: color clock 148
//
//  Since software can change the value of the registers in between clock cycles
//  and phases, it's necessary to develop a mechanism for keeping these registers
//  up to date when doing bulk processing. One method would be to time stamp the
//  register writes, and queue them into a FIFO, but I leave this design decision
//  to the emulator developer. The phase requirements are listed here.
//  AUDC0, AUDC1: used by the pulse generator at both phases
//  AUDF0, AUDF1: used by the clock divider at phase 1
//  AUDV0, AUDV1: used by the volume control at phase 2
//
//  In a real 2600, the volume control is analog, and is affected the instant when
//  AUDVx is written. However, since we generate audio samples at phase 2 of the
//  clock, the granularity of volume control updates is limited to our sample rate,
//  and changes that occur in between clocks result in only the last change prior
//  to phase 2 having an affect on the volume.
//
//-----------------------------------------------------------------------------



/******************************************************************************/
// includes

#include <stdio.h>


/******************************************************************************/
// defines, typedefs, etc.

typedef unsigned int bool;

struct aud_state {
	bool clk_en;
	bool noise_cnt_4;
	bool ext_cnt_hold;
	unsigned int div_cnt;
	unsigned int ext_cnt;
	unsigned int noise_cnt;
};


/******************************************************************************/
// prototypes

void init_aud(struct aud_state *aud0_state, struct aud_state *aud1_state, unsigned short *vol_lut);
bool update_aud_state(struct aud_state *state, unsigned int audf, unsigned int *audc);
unsigned short create_sample(unsigned short *audv_lut, bool aud0, bool aud1, unsigned int audv0, unsigned int audv1);


/******************************************************************************
* Test Bench
*******************************************************************************/
//
//  This test bench is used to exercise the audio logic, and to demonstrate how to
//  call the functions. This code creates the same effect as that created by the 
//  phaser06.bin ROM.
//
int main(void)
{
	int i;
	int scanline;
	unsigned char ram_81;
	unsigned char A;

	bool aud0;						// AUD0 pulse generator state
	bool aud1;						// AUD1 pulse generator state

	unsigned int audf0[2];			// audfx[0]: value at clock1-phase1
	unsigned int audf1[2];			// audfx[1]: value at clock2-phase1
	unsigned int audv0[2];			// audvx[0]: value at clock1-phase2 
	unsigned int audv1[2];			// audvx[1]: value at clock2-phase2
	unsigned int audc0[2][2];		// audcx[0][0]: value at clock1-phase1, audcx[0][1]: value at clock1:phase2
	unsigned int audc1[2][2];		// audcx[1][0]: value at clock2-phase1, audcx[1][1]: value at clock2:phase2

	struct aud_state aud0_state;	// storage for AUD0 state
	struct aud_state aud1_state;	// storage for AUD1 state
	unsigned short vol_lut[256];	// storage for volume LUT
	unsigned short samples[524288]; // audio sample buffer

	FILE *fh;


	i = 0;
	scanline = 0;
	ram_81 = 0x1f;
	init_aud(&aud0_state, &aud1_state, vol_lut);

	while (i < 500000) {

		if (scanline == 0) {
			// audc0 write at color clock 15
			// audc0[0][0] update deferred - write occurred after cycle1:phase1
			audc0[0][1] = 0x06;
			audc0[1][0] = 0x06;
			audc0[1][1] = 0x06;

			// audc1 write at color clock 30
			// audc1[0][0] update deferred - write occurred after cycle1:phase1
			audc1[0][1] = 0x06;
			audc1[1][0] = 0x06;
			audc1[1][1] = 0x06;

		}

		if (scanline == 235) {
			ram_81--;						// LDY ram_81, DEY, STY ram_81
			A = ram_81;						// LDA ram_81
			if (A < 0x80) {					// CMP #$80, BCS skip_EOR
				A ^= 0x7f;					// EOR #$7F
				A = (A << 6) | (A >> 3);	// ROR, ROR, ROR (carry not set)
			}
			else							// ROR, ROR, ROR (carry set)
				A = (A << 6) | 0x20 | (A >> 3);

			// audv0 write at color clock 75
			// audv0[0] update deferred - write occurred after cycle1:phase2
			audv0[1] = A & 0x0f;

			// audv1 write at color clock 84
			// audv1[0] update deferred - write occurred after cycle1:phase2
			audv1[1] = A & 0x0f;

			// audf0 was written at color clock 99, which is after cycle2:phase1 on this line.
		}

		if (scanline == 236) {
			// audf0 write missed phase 1 of both cycles on last line, deferred it to this line.
			audf0[0] = 0x10;
			audf0[1] = 0x10;

			// audf1 write at color clock 15
			// audf1[0] update deferred - write occurred after cycle1:phase1
			audf1[1] = 0x0f;
		}

		if (scanline == 252) {
			// audf1 write at color clock 15
			// audf1[0] update deferred - write occurred after cycle1:phase1
			audf1[1] = 0x10;
		}

		// clock cycle 1
		aud0 = update_aud_state(&aud0_state, audf0[0], audc0[0]);
		aud1 = update_aud_state(&aud1_state, audf1[0], audc1[0]);
		samples[i++] = create_sample(vol_lut, aud0, aud1, audv0[0], audv1[0]);

		// clock cycle 2
		aud0 = update_aud_state(&aud0_state, audf0[1], audc0[1]);
		aud1 = update_aud_state(&aud1_state, audf1[1], audc1[1]);
		samples[i++] = create_sample(vol_lut, aud0, aud1, audv0[1], audv1[1]);

		// end of line, allow deferred updates
		audc0[0][0] = audc0[1][1];	// write occurred after cycle1:phase1
		audc1[0][0] = audc1[1][1];	// write occurred after cycle1:phase1
		audv0[0] = audv0[1];		// write occurred after cycle1:phase2
		audv1[0] = audv1[1];		// write occurred after cycle1:phase2
		audf1[0] = audf1[1];		// write occurred after cycle1:phase1


		if (scanline == 261)
			scanline = 0;
		else
			scanline++;

	}

	fh = fopen("phaser06_31400_signed16_mono.raw", "wb");
	fwrite(samples, 2, i, fh);
	fclose(fh);

}


/******************************************************************************
* Initialize the audio state for each channel
*******************************************************************************/

void init_aud(struct aud_state *aud0_state, struct aud_state *aud1_state, unsigned short *vol_lut)
{
	int i;
	double r1;
	double r2;
	double ra;
	double rb;
	double rc;
	double rd;

	aud0_state->clk_en = 0;
	aud0_state->noise_cnt_4 = 0;
	aud0_state->ext_cnt_hold = 0;
	aud0_state->div_cnt = 0;
	aud0_state->ext_cnt = 0x0f;
	aud0_state->noise_cnt = 0x1f;

	aud1_state->clk_en = 0;
	aud1_state->noise_cnt_4 = 0;
	aud1_state->ext_cnt_hold = 0;
	aud1_state->div_cnt = 0;
	aud1_state->ext_cnt = 0x0f;
	aud1_state->noise_cnt = 0x1f;

	r1 = 1.0;
	ra = 1.0 / 30.0;
	rb = 1.0 / 15.0;
	rc = 1.0 / 7.5;
	rd = 1.0 / 3.75;
	vol_lut[0] = 0;

	for (i = 1; i < 256; i++) {
		r2 = 0.0;

		if (i & 0x01)
			r2 += ra;

		if (i & 0x02)
			r2 += rb;

		if (i & 0x04)
			r2 += rc;

		if (i & 0x08)
			r2 += rd;

		if (i & 0x10)
			r2 += ra;

		if (i & 0x20)
			r2 += rb;

		if (i & 0x40)
			r2 += rc;

		if (i & 0x80)
			r2 += rd;

		r2 = 1.0 / r2;
		vol_lut[i] = (unsigned short) (32768.0 * (1.0 - r2 / (r1 + r2)) + 0.5);
	}
}


/******************************************************************************
* Update the counters, latches, and clock enables
*******************************************************************************/

bool update_aud_state(struct aud_state *state, unsigned int audf, unsigned int *audc)
{

	bool ext_fb;	// extended counter LFSR feedback
	bool noise_fb;	// noise counter LFSR feedback

	// -- Logic updated on phase 1 of the audio clock --
	if (state->clk_en) {
		// latch bit 4 of noise counter
		state->noise_cnt_4 = state->noise_cnt & 0x01;

		// latch extended counter hold condition
		switch (audc[0] & 0x03) {
			case  0x00 :
				state->ext_cnt_hold = 0;
				break;
			case  0x01 :
				state->ext_cnt_hold = 0;
				break;
			case  0x02 :
				state->ext_cnt_hold = ((state->noise_cnt & 0x1e) != 0x02);
				break;
			case  0x03 :
				state->ext_cnt_hold = !state->noise_cnt_4;
		}
	}

	// set (or clear) audio clock enable
	state->clk_en = (state->div_cnt == audf);

	// increment clock divider counter
	if ((state->div_cnt == audf) || (state->div_cnt == 31))
		state->div_cnt = 0;
	else
		state->div_cnt++;


	// -- Logic updated on phase 2 of the audio clock --
	// evaluate noise counter combinitorial logic
	if (state->clk_en) {
		switch (audc[1] & 0x03) {
			case 0x00 :
				noise_fb = ((state->ext_cnt & 0x01) ^ (state->noise_cnt & 0x01)) | !((state->noise_cnt&&1) | (state->ext_cnt != 0x0a)) | !(audc[1] & 0x0c);
				break;
			default   :
				noise_fb = (((state->noise_cnt & 0x04)&&1) ^ (state->noise_cnt & 0x01)) | !state->noise_cnt;
		}

		// evaluate extended counter combinitorial logic
		switch (audc[1] >> 2) {
			case 0x00 :
				ext_fb = (((state->ext_cnt & 0x02)&&1) ^ (state->ext_cnt & 0x01)) & (state->ext_cnt != 0x0a) && (audc[1] & 0x03);
				break;
			case 0x01 :
				ext_fb = !(state->ext_cnt & 0x08);
				break;
			case 0x02 :
				ext_fb = !state->noise_cnt_4;
				break;
			case 0x03 :
				ext_fb = !((state->ext_cnt & 0x02) || !(state->ext_cnt & 0x0e));
		}

		// increment noise counter
		state->noise_cnt >>= 1;
		if (noise_fb)
			state->noise_cnt |= 0x10;

		// increment extended counter
		if (!state->ext_cnt_hold) {
			state->ext_cnt >>= 1;
			if (ext_fb)
				state->ext_cnt |= 0x08;
		}
	}

	// pulse generator output
	return (state->ext_cnt & 0x01);
}


/******************************************************************************
* Generate PCM audio samples
*******************************************************************************/

unsigned short create_sample(unsigned short *audv_lut, bool aud0, bool aud1, unsigned int audv0, unsigned int audv1)
{
	unsigned int lut_index;

	lut_index = 0;
	if (aud0)
		lut_index |= audv0;
	if (aud1)
		lut_index |= (audv1 << 4);
	return audv_lut[lut_index];
}

